MODULE dynzdf_imp_jki
   !!==============================================================================
   !!                    ***  MODULE  dynzdf_imp_jki  ***
   !! Ocean dynamics:  vertical component(s) of the momentum mixing trend
   !!==============================================================================
   !! History : 8.5  !  02-08  (G. Madec)  auto-tasking option
   !!----------------------------------------------------------------------

   !!----------------------------------------------------------------------
   !!   dyn_zdf_imp_jki : update the momentum trend with the vertical
   !!                     diffusion using an implicit time-stepping and
   !!                     j-k-i loops.
   !!----------------------------------------------------------------------
   !! * Modules used
   USE oce             ! ocean dynamics and tracers
   USE dom_oce         ! ocean space and time domain
   USE phycst          ! physical constants
   USE zdf_oce         ! ocean vertical physics
   USE in_out_manager  ! I/O manager
   USE taumod          ! surface ocean stress

   IMPLICIT NONE
   PRIVATE

   !! * Routine accessibility
   PUBLIC dyn_zdf_imp_jki     ! called by step.F90

   !! * Substitutions
#  include "domzgr_substitute.h90"
#  include "vectopt_loop_substitute.h90"
   !!----------------------------------------------------------------------
   !!   OPA 9.0 , LOCEAN-IPSL (2005) 
   !! $Id$
   !! Software governed by the CeCILL licence (modipsl/doc/NEMO_CeCILL.txt)
   !!----------------------------------------------------------------------

CONTAINS

   SUBROUTINE dyn_zdf_imp_jki( kt, p2dt )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE dyn_zdf_imp_jki  ***
      !!                   
      !! ** Purpose :   Compute the trend due to the vert. momentum diffusion
      !!      and the surface forcing, and add it to the general trend of 
      !!      the momentum equations.
      !!
      !! ** Method  :   The vertical momentum mixing trend is given by :
      !!             dz( avmu dz(u) ) = 1/e3u dk+1( avmu/e3uw dk(ua) )
      !!      backward time stepping
      !!      Surface boundary conditions: wind stress input
      !!      Bottom boundary conditions : bottom stress (cf zdfbfr.F)
      !!      Add this trend to the general trend ua :
      !!         ua = ua + dz( avmu dz(u) )
      !!
      !! ** Action : - Update (ua,va) arrays with the after vertical diffusive
      !!               mixing trend.
      !!---------------------------------------------------------------------
      !! * Arguments
      INTEGER , INTENT( in ) ::   kt                          ! ocean time-step index
      REAL(wp), INTENT( in ) ::   p2dt                        ! ocean time-step index

      !! * Local declarations
      INTEGER ::   ji, jj, jk                                 ! dummy loop indices
      INTEGER ::   ikst, ikenm2, ikstp1                       ! temporary integers
      REAL(wp) ::  zrau0r, z2dtf, zcoef, zzws                 ! temporary scalars
      REAL(wp), DIMENSION(jpi,jpk) ::  zwx, zwy, zwz,       & ! workspace
         &                             zwd, zws, zwi, zwt
      !!----------------------------------------------------------------------

      IF( kt == nit000 ) THEN
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) 'dyn_zdf_imp_jki : vertical momentum diffusion implicit operator'
         IF(lwp) WRITE(numout,*) '~~~~~~~~~~~~~~~   auto-task case (j-k-i loop)'
      ENDIF


      ! 0. Local constant initialization
      ! --------------------------------
      zrau0r = 1. / rau0      ! inverse of the reference density

      !                                                ! ===============
      DO jj = 2, jpjm1                                 !  Vertical slab
         !                                             ! ===============
         ! 1. Vertical diffusion on u
         ! ---------------------------

         ! Matrix and second member construction
         ! bottom boundary condition: only zws must be masked as avmu can take
         ! non zero value at the ocean bottom depending on the bottom friction
         ! used (see zdfmix.F)
         DO jk = 1, jpkm1
            DO ji = 2, jpim1
               zcoef = - p2dt / fse3u(ji,jj,jk)
               zwi(ji,jk) = zcoef * avmu(ji,jj,jk  ) / fse3uw(ji,jj,jk  )
               zzws       = zcoef * avmu(ji,jj,jk+1) / fse3uw(ji,jj,jk+1)
               zws(ji,jk) = zzws * umask(ji,jj,jk+1)
               zwd(ji,jk) = 1. - zwi(ji,jk) - zzws
               zwy(ji,jk) = ub(ji,jj,jk) + p2dt * ua(ji,jj,jk)
            END DO
         END DO

         ! Surface boudary conditions
         DO ji = 2, jpim1
            z2dtf = p2dt / ( fse3u(ji,jj,1)*rau0 )
            zwi(ji,1) = 0.
            zwd(ji,1) = 1. - zws(ji,1)
            zwy(ji,1) = zwy(ji,1) + z2dtf * taux(ji,jj)
         END DO

         ! Matrix inversion starting from the first level
         ikst = 1
!!----------------------------------------------------------------------
!!         ZDF.MATRIXSOLVER
!!       ********************
!!----------------------------------------------------------------------
!! Matrix inversion
!   solve m.x = y  where m is a tri diagonal matrix ( jpk*jpk )
!
!        ( zwd1 zws1   0    0    0  )( zwx1 ) ( zwy1 )
!        ( zwi2 zwd2 zws2   0    0  )( zwx2 ) ( zwy2 )
!        (  0   zwi3 zwd3 zws3   0  )( zwx3 )=( zwy3 )
!        (        ...               )( ...  ) ( ...  )
!        (  0    0    0   zwik zwdk )( zwxk ) ( zwyk )
!
!   m is decomposed in the product of an upper and lower triangular
!   matrix
!   The 3 diagonal terms are in 2d arrays: zwd, zws, zwi
!   The second member is in 2d array zwy
!   The solution is in 2d array zwx
!   The 2d arry zwt and zwz are work space arrays
!
!   N.B. the starting vertical index (ikst) is equal to 1 except for
!   the resolution of tke matrix where surface tke value is prescribed
!   so that ikstrt=2.
!!----------------------------------------------------------------------

         ikstp1 = ikst + 1
         ikenm2 = jpk - 2
         DO ji = 2, jpim1
            zwt(ji,ikst) = zwd(ji,ikst)
         END DO
         DO jk = ikstp1, jpkm1
            DO ji = 2, jpim1
               zwt(ji,jk) = zwd(ji,jk) - zwi(ji,jk) * zws(ji,jk-1) / zwt(ji,jk-1)
            END DO
         END DO
         DO ji = 2, jpim1
            zwz(ji,ikst) = zwy(ji,ikst)
         END DO
         DO jk = ikstp1, jpkm1
            DO ji = 2, jpim1
               zwz(ji,jk) = zwy(ji,jk) - zwi(ji,jk) / zwt(ji,jk-1) * zwz(ji,jk-1)
            END DO
         END DO
         DO ji = 2, jpim1
            zwx(ji,jpkm1) = zwz(ji,jpkm1) / zwt(ji,jpkm1)
         END DO
         DO jk = ikenm2, ikst, -1
            DO ji = 2, jpim1
            zwx(ji,jk) =( zwz(ji,jk) - zws(ji,jk) * zwx(ji,jk+1) ) / zwt(ji,jk)
            END DO
         END DO
      
         ! Normalization to obtain the general momentum trend ua
         DO jk = 1, jpkm1
            DO ji = 2, jpim1
               ua(ji,jj,jk) = ( zwx(ji,jk) - ub(ji,jj,jk) ) / p2dt
            END DO
         END DO


         ! 2. Vertical diffusion on v
         ! ---------------------------
         ! Matrix and second member construction
         ! bottom boundary condition: only zws must be masked as avmv can take
         ! non zero value at the ocean bottom depending on the bottom friction
         ! used (see zdfmix.F)
         DO jk = 1, jpkm1
            DO ji = 2, jpim1
               zcoef = -p2dt/fse3v(ji,jj,jk)
               zwi(ji,jk) = zcoef * avmv(ji,jj,jk  ) / fse3vw(ji,jj,jk  )
               zzws       = zcoef * avmv(ji,jj,jk+1) / fse3vw(ji,jj,jk+1)
               zws(ji,jk) =  zzws * vmask(ji,jj,jk+1)
               zwd(ji,jk) = 1. - zwi(ji,jk) - zzws
               zwy(ji,jk) = vb(ji,jj,jk) + p2dt * va(ji,jj,jk)
            END DO
         END DO

         ! Surface boudary conditions
         DO ji = 2, jpim1
            z2dtf = p2dt / ( fse3v(ji,jj,1)*rau0 )
            zwi(ji,1) = 0.e0
            zwd(ji,1) = 1. - zws(ji,1)
            zwy(ji,1) = zwy(ji,1) + z2dtf * tauy(ji,jj)
         END DO

         ! Matrix inversion starting from the first level
         ikst = 1
!!----------------------------------------------------------------------
!!         ZDF.MATRIXSOLVER
!!       ********************
!!----------------------------------------------------------------------
!! Matrix inversion
!   solve m.x = y  where m is a tri diagonal matrix ( jpk*jpk )
!
!        ( zwd1 zws1   0    0    0  )( zwx1 ) ( zwy1 )
!        ( zwi2 zwd2 zws2   0    0  )( zwx2 ) ( zwy2 )
!        (  0   zwi3 zwd3 zws3   0  )( zwx3 )=( zwy3 )
!        (        ...               )( ...  ) ( ...  )
!        (  0    0    0   zwik zwdk )( zwxk ) ( zwyk )
!
!   m is decomposed in the product of an upper and lower triangular
!   matrix
!   The 3 diagonal terms are in 2d arrays: zwd, zws, zwi
!   The second member is in 2d array zwy
!   The solution is in 2d array zwx
!   The 2d arry zwt and zwz are work space arrays
!
!   N.B. the starting vertical index (ikst) is equal to 1 except for
!   the resolution of tke matrix where surface tke value is prescribed
!   so that ikstrt=2.
!!----------------------------------------------------------------------

         ikstp1 = ikst + 1
         ikenm2 = jpk - 2
         DO ji = 2, jpim1
            zwt(ji,ikst) = zwd(ji,ikst)
         END DO
         DO jk = ikstp1, jpkm1
            DO ji = 2, jpim1
               zwt(ji,jk) = zwd(ji,jk) - zwi(ji,jk) * zws(ji,jk-1) / zwt(ji,jk-1)
            END DO
         END DO
         DO ji = 2, jpim1
            zwz(ji,ikst) = zwy(ji,ikst)
         END DO
         DO jk = ikstp1, jpkm1
            DO ji = 2, jpim1
               zwz(ji,jk) = zwy(ji,jk) - zwi(ji,jk) / zwt(ji,jk-1) * zwz(ji,jk-1)
            END DO
         END DO
         DO ji = 2, jpim1
            zwx(ji,jpkm1) = zwz(ji,jpkm1) / zwt(ji,jpkm1)
         END DO
         DO jk = ikenm2, ikst, -1
            DO ji = 2, jpim1
               zwx(ji,jk) =( zwz(ji,jk) - zws(ji,jk) * zwx(ji,jk+1) ) / zwt(ji,jk)
            END DO
         END DO
      
         ! Normalization to obtain the general momentum trend va
         DO jk = 1, jpkm1
            DO ji = 2, jpim1
               va(ji,jj,jk) = ( zwx(ji,jk) - vb(ji,jj,jk) ) / p2dt
            END DO
         END DO
         !                                             ! ===============
      END DO                                           !   End of slab
      !                                                ! ===============

   END SUBROUTINE dyn_zdf_imp_jki

   !!==============================================================================
END MODULE dynzdf_imp_jki
